# -*- coding: utf-8 -*-
import warnings
from relevanceai import vector_tools
from relevanceai.interfaces import Client

# Cluster _Base Utilities
from relevanceai.workflows.cluster_ops.cluster_base import (
    ClusterBase,
    CentroidClusterBase,
)
from relevanceai.workflows.cluster_ops.clusterops import ClusterOps
from relevanceai.workflows.dim_reduction_ops.dim_reduction_ops import (
    ReduceDimensionsOps,
)

# Fix the name
from relevanceai.package_utils import datasets
from relevanceai.package_utils.datasets import mock_documents
import requests

# Import useful utility if possible as well
try:
    from jsonshower import show_json
except ModuleNotFoundError:
    pass

__version__ = "2.0.0"

try:
    pypi_data = requests.get("https://pypi.org/pypi/relevanceai/json").json()
    pypi_info = pypi_data.get("info", None)
    if pypi_info is not None:
        latest_version = pypi_info.get("version", None)
    else:
        latest_version = None

    if __version__ != latest_version and latest_version is not None:
        changelog_url: str = (
            f"https://relevanceai.readthedocs.io/en/{__version__}/changelog.html"
        )
        MESSAGE = """We noticed you don't have the latest version!
We recommend updating to the latest version ({latest_version}) to get all bug fixes and newest features!
You can do this by running pip install -U relevanceai.
Changelog: {changelog_url}.""".format(  # type: ignore
            version=__version__,
            latest_version=latest_version,
            changelog_url=changelog_url,
        )
        warnings.warn(MESSAGE)
except:
    pass



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import warnings
from typing import List, Optional
from tqdm.auto import tqdm
from relevanceai.package_utils.analytics_funcs import track
from relevanceai.dataset.crud.dataset_write import Write
from relevanceai.package_utils.logger import FileLogger
from relevanceai.package_utils.version_decorators import introduced_in_version, beta


class Cluster(Write):
    @track
    def cluster(self, model, alias, vector_fields, **kwargs):
        """
        Performs KMeans Clustering on over a vector field within the dataset.

        .. warning::
            Deprecated in v0.33 in favour of df.auto_cluster.

        Parameters
        ------------
        model : Class
            The clustering model to use
        vector_fields : str
            The vector fields over which to cluster

        Example
        ------------
        .. code-block::

            from relevanceai import Client
            from relevanceai ClusterOps
            from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            vector_field = "vector_field_"
            n_clusters = 10

            model = KMeansModel(k=n_clusters)

            df.cluster(model=model, alias=f"kmeans-{n_clusters}", vector_fields=[vector_field])
        """
        from relevanceai.workflows.cluster_ops.clusterops import ClusterOps

        clusterer = ClusterOps(
            model=model,
            alias=alias,
            api_key=self.api_key,
            project=self.project,
            firebase_uid=self.firebase_uid,
        )
        clusterer.fit_predict_update(dataset=self, vector_fields=vector_fields)
        return clusterer

    def cluster_keyphrases(
        self,
        vector_fields: List[str],
        text_fields: List[str],
        cluster_alias: str,
        cluster_field: str = "_cluster_",
        num_clusters: int = 100,
        most_common: int = 10,
        preprocess_hooks: Optional[List[callable]] = None,
        algorithm: str = "rake",
        n: int = 2,
    ):
        """
        Simple implementation of the cluster word cloud.

        Parameters
        ------------
        vector_fields: list
            The list of vector fields
        text_fields: list
            The list of text fields
        cluster_alias: str
            The alias of the cluster
        cluster_field: str
            The cluster field to try things on
        num_clusters: int
            The number of clusters
        preprocess_hooks: list
            The preprocess hooks
        algorithm: str
            The algorithm to use
        n: int
            The number of words

        """
        preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

        vector_fields_str = ".".join(sorted(vector_fields))
        field = f"{cluster_field}.{vector_fields_str}.{cluster_alias}"
        all_clusters = self.facets([field], page_size=num_clusters)
        cluster_counters = {}
        if "results" in all_clusters:
            all_clusters = all_clusters["results"]
        # TODO: Switch to multiprocessing
        for c in tqdm(all_clusters[field]):
            cluster_value = c[field]
            top_words = self.keyphrases(
                text_fields=text_fields,
                n=n,
                filters=[
                    {
                        "field": field,
                        "filter_type": "contains",
                        "condition": "==",
                        "condition_value": cluster_value,
                    }
                ],
                most_common=most_common,
                preprocess_hooks=preprocess_hooks,
                algorithm=algorithm,
            )
            cluster_counters[cluster_value] = top_words
        return cluster_counters

    # TODO: Add keyphrases to auto cluster
    # def auto_cluster_keyphrases(
    #     vector_fields: List[str],
    #     text_fields: List[str],
    #     cluster_alias: str,
    #     deployable_id: str,
    #     n: int = 2,
    #     cluster_field: str = "_cluster_",
    #     num_clusters: int = 100,
    #     preprocess_hooks: Optional[List[callable]] = None,
    # ):
    #     """
    #     # TODO:
    #     """
    #     pass

    def _add_cluster_word_cloud_to_config(self, data, cluster_value, top_words):
        # TODO: Add this to wordcloud deployable
        # hacky way I implemented to add top words to config
        data["configuration"]["cluster-labels"][cluster_value] = ", ".join(
            [k for k in top_words if k != "machine learning"]
        )
        data["configuration"]["cluster-descriptions"][cluster_value] = str(top_words)

    @track
    def auto_cluster(
        self,
        alias: str,
        vector_fields: List[str],
        model=None,
        chunksize: int = 1024,
        filters: Optional[list] = None,
        parent_alias: Optional[str] = None,
    ):
        """
        Automatically cluster in 1 line of code.
        It will retrieve documents, run fitting on the documents and then
        update the database.
        There are only 2 supported clustering algorithms at the moment:
        - kmeans
        - minibatchkmeans

        In order to choose the number of clusters, simply add a number
        after the dash like `kmeans-8` or `minibatchkmeans-50`.

        Under the hood, it uses scikit learn defaults or best practices.

        This returns a ClusterOps object and is a wrapper on top of
        `ClusterOps`.

        Parameters
        ----------
        alias : str
            The clustering model (as a str) to use and n_clusters. Delivered in a string separated by a '-'
            Supported aliases at the moment are 'kmeans','kmeans-10', 'kmeans-X' (where X is a number), 'minibatchkmeans',
                'minibatchkmeans-10', 'minibatchkmeans-X' (where X is a number)
        vector_fields : List
            A list vector fields over which to cluster

        Example
        ----------

        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            # run kmeans with default 10 clusters
            clusterer = df.auto_cluster("kmeans", vector_fields=[vector_field])
            clusterer.list_closest_to_center()

            # Run k means clustering with 8 clusters
            clusterer = df.auto_cluster("kmeans-8", vector_fields=[vector_field])

            # Run minibatch k means clustering with 8 clusters
            clusterer = df.auto_cluster("minibatchkmeans-8", vector_fields=[vector_field])

            # Run minibatch k means clustering with 20 clusters
            clusterer = df.auto_cluster("minibatchkmeans-20", vector_fields=[vector_field])

        You can alternatively run this using kmeans.

        .. code-block::

            from relevanceai import Client

            client = Client()

            from relevanceai.package_utils.datasets import mock_documents

            ds = client.Dataset('sample')
            ds.upsert_documents(mock_documents(100))
            # Run initial kmeans to get clusters
            ds.auto_cluster('kmeans-3', vector_fields=["sample_1_vector_"])
            # Run separate K Means to get subclusters
            cluster_ops = ds.auto_cluster(
                'kmeans-2',
                vector_fields=["sample_1_vector_"],
                parent_alias="kmeans-3"
            )


        """
        if model is None:
            return self._auto_cluster_string(
                alias=alias,
                vector_fields=vector_fields,
                chunksize=1024,
                filters=filters,
                parent_alias=parent_alias,
            )
        else:
            return self._auto_cluster_model(
                alias=alias,
                model=model,
                vector_fields=vector_fields,
                chunksize=chunksize,
                filters=filters,
                parent_alias=parent_alias,
            )

    def _auto_cluster_model(
        self,
        alias: str,
        vector_fields: List[str],
        chunksize: int = 1024,
        filters: Optional[list] = None,
        parent_alias: Optional[str] = None,
        model=None,
    ):
        filters = [] if filters is None else filters

        cluster_args = alias.split("-")
        algorithm = cluster_args[0]
        if len(cluster_args) > 1:
            n_clusters = int(cluster_args[1])
        else:
            print("No clusters are detected, defaulting to 8")
            n_clusters = 8
        if n_clusters >= chunksize:
            raise ValueError("Number of clustesr exceed chunksize.")

        num_documents = self.get_number_of_documents(self.dataset_id)

        if num_documents <= n_clusters:
            warnings.warn(
                "You seem to have more clusters than documents. We recommend reducing the number of clusters."
            )

        from relevanceai.workflows.cluster_ops.clusterops import ClusterOps

        clusterer: ClusterOps = ClusterOps(
            model=model,
            alias=alias,
            api_key=self.api_key,
            project=self.project,
            firebase_uid=self.firebase_uid,
            dataset_id=self.dataset_id,
            vector_fields=vector_fields,
            parent_alias=parent_alias,
        )
        if parent_alias:
            clusterer.subcluster_predict_update(
                dataset=self,
                vector_fields=vector_fields,
                filters=filters,
            )
        else:
            clusterer.fit_predict_update(
                dataset=self,
                vector_fields=vector_fields,
                include_grade=True,
                filters=filters,
            )
        return clusterer

    def _store_subcluster_metadata(
        self, vector_fields: list, alias: str, parent_alias: str
    ):
        # Store metadata around subclustering
        field = str("-".join(vector_fields)) + "." + alias
        metadata = self.metadata
        if "subcluster" not in metadata:
            metadata["subclusters"] = []
        metadata["subclusters"].append(
            {
                "parent_alias": parent_alias,
                "alias": alias,
                "vector_fields": vector_fields,
            }
        )
        self.upsert_metadata(metadata)

    @track
    def _auto_cluster_string(
        self,
        alias: str,
        vector_fields: List[str],
        chunksize: int = 1024,
        filters: Optional[list] = None,
        parent_alias: Optional[str] = None,
    ):
        """
        Automatically cluster in 1 line of code.
        It will retrieve documents, run fitting on the documents and then
        update the database.
        There are only 2 supported clustering algorithms at the moment:
        - kmeans
        - minibatchkmeans

        In order to choose the number of clusters, simply add a number
        after the dash like `kmeans-8` or `minibatchkmeans-50`.

        Under the hood, it uses scikit learn defaults or best practices.

        This returns a ClusterOps object and is a wrapper on top of
        `ClusterOps`.

        Parameters
        ----------
        alias : str
            The clustering model (as a str) to use and n_clusters. Delivered in a string separated by a '-'
            Supported aliases at the moment are 'kmeans','kmeans-10', 'kmeans-X' (where X is a number), 'minibatchkmeans',
                'minibatchkmeans-10', 'minibatchkmeans-X' (where X is a number)
        vector_fields : List
            A list vector fields over which to cluster

        Example
        ----------

        .. code-block::

            from relevanceai import Client

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            # run kmeans with default 10 clusters
            clusterer = df.auto_cluster("kmeans", vector_fields=[vector_field])
            clusterer.list_closest_to_center()

            # Run k means clustering with 8 clusters
            clusterer = df.auto_cluster("kmeans-8", vector_fields=[vector_field])

            # Run minibatch k means clustering with 8 clusters
            clusterer = df.auto_cluster("minibatchkmeans-8", vector_fields=[vector_field])

            # Run minibatch k means clustering with 20 clusters
            clusterer = df.auto_cluster("minibatchkmeans-20", vector_fields=[vector_field])

        You can alternatively run this using kmeans.

        .. code-block::

            from relevanceai import Client

            client = Client()

            from relevanceai.package_utils.datasets import mock_documents

            ds = client.Dataset('sample')
            ds.upsert_documents(mock_documents(100))
            # Run initial kmeans to get clusters
            ds.auto_cluster('kmeans-3', vector_fields=["sample_1_vector_"])
            # Run separate K Means to get subclusters
            cluster_ops = ds.auto_cluster(
                'kmeans-2',
                vector_fields=["sample_1_vector_"],
                parent_alias="kmeans-3"
            )


        """
        filters = [] if filters is None else filters

        cluster_args = alias.split("-")
        algorithm = cluster_args[0]
        if len(cluster_args) > 1:
            n_clusters = int(cluster_args[1])
        else:
            print("No clusters are detected, defaulting to 8")
            n_clusters = 8
        if n_clusters >= chunksize:
            raise ValueError("Number of clustesr exceed chunksize.")

        num_documents = self.get_number_of_documents(self.dataset_id)

        if num_documents <= n_clusters:
            warnings.warn(
                "You seem to have more clusters than documents. We recommend reducing the number of clusters."
            )

        from relevanceai.workflows.cluster_ops.clusterops import ClusterOps

        if algorithm.lower() == "kmeans":
            from sklearn.cluster import KMeans

            model = KMeans(n_clusters=n_clusters)
            clusterer: ClusterOps = ClusterOps(
                model=model,
                alias=alias,
                api_key=self.api_key,
                project=self.project,
                firebase_uid=self.firebase_uid,
                dataset_id=self.dataset_id,
                vector_fields=vector_fields,
                parent_alias=parent_alias,
            )
            if parent_alias:
                clusterer.subcluster_predict_update(
                    dataset=self,
                    vector_fields=vector_fields,
                    filters=filters,
                )

                self._store_subcluster_metadata(
                    vector_fields=vector_fields, alias=alias, parent_alias=parent_alias
                )
            else:
                clusterer.fit_predict_update(
                    dataset=self,
                    vector_fields=vector_fields,
                    include_grade=True,
                    filters=filters,
                )

        elif algorithm.lower() == "hdbscan":
            raise ValueError(
                "HDBSCAN is soon to be released as an alternative clustering algorithm"
            )
        elif algorithm.lower() == "minibatchkmeans":
            from sklearn.cluster import MiniBatchKMeans

            model = MiniBatchKMeans(n_clusters=n_clusters)

            clusterer = ClusterOps(
                model=model,
                alias=alias,
                api_key=self.api_key,
                project=self.project,
                firebase_uid=self.firebase_uid,
                dataset_id=self.dataset_id,
                vector_fields=vector_fields,
                parent_alias=parent_alias,
            )

            if parent_alias:
                print("subpartial fit...")
                clusterer.subpartialfit_predict_update(
                    dataset=self,
                    vector_fields=vector_fields,
                    filters=filters,
                )

                self._store_subcluster_metadata(
                    vector_fields=vector_fields, alias=alias, parent_alias=parent_alias
                )

            else:
                clusterer.partial_fit_predict_update(
                    dataset=self,
                    vector_fields=vector_fields,
                    chunksize=chunksize,
                    filters=filters,
                )
        else:
            raise ValueError("Only KMeans clustering is supported at the moment.")

        return clusterer

    def auto_text_cluster_dashboard(
        self,
        text_fields: List[str],
        alias: str,
        chunksize: int = 1024,
        filters: Optional[list] = None,
        text_encoder=None,
    ):
        """
        Convenient way to vectorize and cluster text fields.

        Parameters
        ----------
        text_fields: List[str]
            A list of text fields to vectorize and cluster

        alias: str
            The name of the clustring application. The alias is required to
            be of the form "{algorithm}-{n_clusters}" where:
                * algorithm is the clustering algorithm to be used; and
                * n_clusters is the number of clusters

        chunksize: int
            The size of the chunks

        filters: Optional[list]
            A list of filters to apply over the fields to vectorize

        text_encoder:
            A deep learning text encoder from the vectorhub library. If no
            encoder is specified, a default encoder (USE2Vec) is loaded.

        Returns
        -------
        A dictionary indicating the outcome status

        Example
        -------

        .. code-block::

            from relevanceai import Client

            client = Client()

            ds = client.Dataset("sample_dataset_id")

            ds.auto_text_cluster_dashboard(text_fields=["sample_text_field"])

        """
        filters = [] if filters is None else filters

        fields = []
        for field in text_fields:
            try:
                if not "text" == self.schema[field]:
                    fields.append(field)
            except KeyError:
                raise KeyError(f"'{field}' is an invalid field")
        else:
            if fields:
                raise ValueError(
                    "The following fields are not text fields: " f"{', '.join(fields)}"
                )

        results = self.vectorize(text_fields=text_fields, text_encoder=text_encoder)
        if "added_vectors" not in results:
            # If there were errors in vectorizing, then quit immediately and return errors
            return results

        new_vectors = results["added_vectors"]
        existing_vectors = results["skipped_vectors"]

        self.auto_cluster(
            alias=alias,
            vector_fields=new_vectors + existing_vectors,
            chunksize=chunksize,
            filters=filters,
        )

        print(
            "Build your clustering app here: "
            f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
        )

    @beta
    def launch_cluster_app(self, configuration: dict):
        """
        Launch an app with a given configuration
        """
        if "configuration" in configuration:
            configuration = configuration["configuration"]
        results = self.deployables.create(
            dataset_id=self.dataset_id, configuration=configuration
        )

        # After you have created an app
        url = f"https://cloud.relevance.ai/dataset/{results['dataset_id']}/deploy/cluster/{self.project}/{self.api_key}/{results['deployable_id']}/{self.region}"
        print(f"You can now access your deployable at {url}.")
        return url



"""
Dimensionality Reduction Ops

.. warning::
    This is a beta feature and is currently in development.

Reducing dimensions for just documents.

.. code-block::

    from relevanceai import Client 
    client = Client()

    from relevanceai.package_utils.datasets import mock_documents
    docs = mock_documents(10)

    from relevanceai.dim_reduction_ops import ReduceDimensionsOps
    from sklearn.decomposition import PCA
    model = PCA(n_components=2)
    dim_reducer = ReduceDimensionsOps(model)
    dim_reducer.reduce_dimensions(fields=["sample_1_vector_"], documents=documents)

"""
import copy
from doc_utils import DocUtils


class ReduceDimensionsOps(DocUtils):
    def __init__(self, model, alias: str):
        """
        Dim Reduction Ops

        Parameters
        --------------

        model
            The model to run dimensionality reduction. This requires a `fit_transform` method.
        alias: str
            The alias of the dimensionality reduction
        """
        self.model = model
        if not hasattr(model, "fit_transform"):
            raise AttributeError("‼️ Model needs to have a fit_transform method.")
        self.alias = alias

    def reduce_dimensions(self, fields: list, documents: list, inplace: bool = True):
        """
        Reduce Dimensions

        Parameters
        --------------

        fields: list
            The list of fields to run dimensionality reduction on. Currently
            only supports 1 field.
        documents: list
            The list of documents to run dimensionality reduction on
        inplace: bool
            If True, replaces the original documents, otherwise it returns
            a new set of documents with only the dr vectors in it and the _id
        """
        self.fields = fields
        self.documents = documents
        if len(fields) == 1:
            values = self.get_field_across_documents(fields[0], documents)
        else:
            raise ValueError("Supporting multiple fields not supported yet.")
        dr_values = self.model.fit_transform(values)
        new_documents = self.set_dr_field_across_documents(
            fields, dr_values, documents, inplace=inplace
        )
        return new_documents

    def set_dr_field_across_documents(
        self, fields: list, values: list, documents: list, inplace: bool = True
    ):
        """
        Setting the DR field allows users to quickly follow Relevance AI
        best practice in naming.
        It follows `_dr_.*field_name*.alias`

        Parameters
        -------------

        fields:
            The list of fields to set
        values:
            The list of values to set
        documents:
            List of documents to set

        """
        fields_joined = ".".join(fields)
        field_name = f"_dr_.{fields_joined}.{self.alias}"

        if inplace:
            self.set_field_across_documents(field_name, values, documents)
            return documents
        new_documents = [{"_id": d["_id"]} for d in documents]
        self.set_field_across_documents(field_name, values, new_documents)
        return new_documents



import pytest


def test_get_sample_ecommerce_dataset():
    from relevanceai.package_utils.datasets import get_ecommerce_1_dataset

    assert len(get_ecommerce_1_dataset(number_of_documents=100)) == 100



import pandas as pd
from typing import Dict, List, Optional, Union

from relevanceai.package_utils.analytics_funcs import track
from relevanceai.dataset.export.interface import Export
from relevanceai.dataset.auto.dataset_operations import Operations
from relevanceai.dataset.crud.dataset_series import Series
from relevanceai.dataset.search.search import Search
from relevanceai.dataset.vis.plot import Plot

# TODO: Add game dataset
_GLOBAL_DATASETS = ["_mock_dataset_"]


def str2bool(v: str):
    return v.lower() in ("yes", "true", "t", "1")


class Dataset(Export, Plot, Operations, Search):
    @track
    def __init__(
        self,
        project: str,
        api_key: str,
        dataset_id: str,
        firebase_uid: str,
        fields: Optional[list] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, List]] = None,
        text_fields: Optional[List[str]] = None,
        **kw,
    ):
        self.project = project
        self.api_key = api_key
        self.firebase_uid = firebase_uid
        self.fields = [] if fields is None else fields
        self.dataset_id = dataset_id
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        self.text_fields = [] if text_fields is None else text_fields
        self.firebase_uid = firebase_uid

        super().__init__(
            project=project,
            api_key=api_key,
            firebase_uid=firebase_uid,
            fields=fields,
            dataset_id=dataset_id,
            image_fields=image_fields,
            audio_fields=audio_fields,
            highlight_fields=highlight_fields,
            text_fields=text_fields,
            **kw,
        )
        self.search = Search(
            project=project,
            api_key=api_key,
            firebase_uid=firebase_uid,
        )
        # add global datasets
        if self.dataset_id in _GLOBAL_DATASETS:
            # avoid re-inserting if it already exists
            if self.dataset_id not in self.datasets.list()["datasets"]:
                from relevanceai.package_utils.datasets import mock_documents
                from relevanceai.package_utils.analytics_funcs import fire_and_forget

                @fire_and_forget
                def add_mock_dataset():
                    self.upsert_documents(mock_documents(100))

                add_mock_dataset()

    # def __getattr__(self, attr):
    #     if hasattr(pd.DataFrame, attr):
    #         df = self.to_pandas_dataframe(show_progress_bar=True)
    #         try:
    #             return getattr(df, attr)
    #         except SyntaxError:
    #             raise AttributeError(f"'{attr}' is an invalid attribute")
    #     raise AttributeError(f"'{attr}' is an invalid attribute")

    @track
    def __getitem__(self, field: Union[List[str], str]):
        """
        Returns a Series Object that selects a particular field within a dataset
        Parameters
        ----------
        field: Union[List, str]
            The particular field within the dataset
        Returns
        -------
        Tuple
            (N, C)
        Example
        ---------------
        .. code-block::
            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            series = df[field]
        """
        if isinstance(field, str):
            return Series(
                project=self.project,
                api_key=self.api_key,
                dataset_id=self.dataset_id,
                firebase_uid=self.firebase_uid,
                field=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        elif isinstance(field, list):
            return Dataset(
                project=self.project,
                api_key=self.api_key,
                dataset_id=self.dataset_id,
                firebase_uid=self.firebase_uid,
                fields=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        else:
            raise TypeError("Field needs to be a list or a string.")

    @track
    def launch_search_app(self):
        """
        Launches the link to the search application to start building
        """
        return (
            f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/search"
        )



"""
ClusterOps class to run clustering. It is intended to be integrated with
models that inherit from `ClusterBase`.

You can run the ClusterOps as such:

.. code-block::

    from relevanceai import Client
    client = Client()
    df = client.Dataset("sample_dataset")

    from sklearn.cluster import KMeans
    model = KMeans(n_clusters=2)
    cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
    cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

You can view other examples of how to interact with this class here :ref:`integration`.

"""
import os
import json
import getpass
import warnings

import numpy as np

from relevanceai.api.client import BatchAPIClient
from typing import Union, List, Dict, Optional, Callable, Set
from relevanceai.workflows.cluster_ops.cluster_base import (
    ClusterBase,
    CentroidClusterBase,
    BatchClusterBase,
    HDBSCANClusterBase,
    SklearnCentroidBase,
)
from relevanceai.workflows.cluster_ops.cluster_groupby import ClusterGroupby, ClusterAgg

from relevanceai.package_utils.analytics_funcs import track

# We use the second import because the first one seems to be causing errors with isinstance
# from relevanceai.dataset.crud import Dataset
from relevanceai.package_utils.integration_checks import (
    is_sklearn_available,
    is_hdbscan_available,
)
from relevanceai.dataset_interface import Dataset

from relevanceai.package_utils.errors import NoDocumentsError
from relevanceai.package_utils.version_decorators import beta
from relevanceai.package_utils.concurrency import multiprocess
from relevanceai.workflows.cluster_ops.cluster_evaluate import ClusterEvaluate
from doc_utils import DocUtils


from tqdm.auto import tqdm

SILHOUETTE_INFO = """
Good clusters have clusters which are highly seperated and elements within which are highly cohesive. <br/>
<b>Silohuette Score</b> is a metric from <b>-1 to 1</b> that calculates the average cohesion and seperation of each element, with <b>1</b> being clustered perfectly, <b>0</b> being indifferent and <b>-1</b> being clustered the wrong way"""

RANDOM_INFO = """Good clusters have elements, which, when paired, belong to the same cluster label and same ground truth label. <br/>
<b>Rand Index</b> is a metric from <b>0 to 1</b> that represents the percentage of element pairs that have a matching cluster and ground truth labels with <b>1</b> matching perfect and <b>0</b> matching randomly. <br/> <i>Note: This measure is adjusted for randomness so does not equal the exact numerical percentage.</i>"""

HOMOGENEITY_INFO = """Good clusters only have elements from the same ground truth within the same cluster<br/>
<b>Homogeneity</b> is a metric from <b>0 to 1</b> that represents whether clusters contain only elements in the same ground truth with <b>1</b> being perfect and <b>0</b> being absolutely incorrect."""

COMPLETENESS_INFO = """Good clusters have all elements from the same ground truth within the same cluster <br/>
<b>Completeness</b> is a metric from <b>0 to 1</b> that represents whether clusters contain all elements in the same ground truth with <b>1</b> being perfect and <b>0</b> being absolutely incorrect."""

AVERAGE_SCORE = """Averages other metrics by first normalising values between 0 and 1 <br/>
<b>Average</b> is a metric from <b>0 to 1</b> that averages other metrics by first normalising values between 0 and 1."""

METRIC_DESCRIPTION = {
    "silhouette": SILHOUETTE_INFO,
    "random": RANDOM_INFO,
    "homogeneity": HOMOGENEITY_INFO,
    "completeness": COMPLETENESS_INFO,
    "average": AVERAGE_SCORE,
}


class ClusterOps(ClusterEvaluate):

    _cred_fn = ".creds.json"

    """
    ClusterOps class allows users to set up any clustering model to fit on a Dataset.

    You can read about the other parameters here: https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html

    Example
    -----------

    .. code-block::

        from relevanceai import Client
        client = Client()
        df = client.Dataset("sample_dataset")

        from sklearn.cluster import KMeans
        model = KMeans(n_clusters=2)
        cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
        cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

    """

    def __init__(
        self,
        alias: str,
        project: str,
        api_key: str,
        firebase_uid: str,
        model: Union[BatchClusterBase, ClusterBase, CentroidClusterBase] = None,
        dataset_id: Optional[str] = None,
        vector_fields: Optional[List[str]] = None,
        cluster_field: str = "_cluster_",
        parent_alias: str = None,
    ):
        self.alias = alias
        self.parent_alias = parent_alias
        self.cluster_field = cluster_field
        if model is None:
            warnings.warn(
                "No model is specified, you will not be able to train a clustering algorithm."
            )

        self.model = self._assign_model(model)
        self.firebase_uid = firebase_uid

        if dataset_id is not None:
            self.dataset_id: str = dataset_id
        if vector_fields is not None:
            self.vector_fields = vector_fields

        if project is None or api_key is None:
            project, api_key = self._token_to_auth()
        else:
            self.project: str = project
            self.api_key: str = api_key

        super().__init__(project=project, api_key=api_key, firebase_uid=firebase_uid)

    def __call__(self):
        self.groupby = ClusterGroupby(
            project=self.project,
            api_key=self.api_key,
            dataset_id=self.dataset_id,
            firebase_uid=self.firebase_uid,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )
        self.agg = ClusterAgg(
            project=self.project,
            api_key=self.api_key,
            dataset_id=self.dataset_id,
            firebase_uid=self.firebase_uid,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    @track
    def groupby(self):
        return ClusterGroupby(
            project=self.project,
            api_key=self.api_key,
            dataset_id=self.dataset_id,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )

    @track
    def agg(self, groupby_call):
        """Aggregate the cluster class."""
        self.groupby = ClusterGroupby(
            project=self.project,
            api_key=self.api_key,
            dataset_id=self.dataset_id,
            firebase_uid=self.firebase_uid,
            alias=self.alias,
            vector_fields=self.vector_fields,
        )
        return ClusterAgg(
            project=self.project,
            api_key=self.api_key,
            dataset_id=self.dataset_id,
            firebase_uid=self.firebase_uid,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    # Adding first-class sklearn integration
    def _assign_sklearn_model(self, model):
        # Add support for not just sklearn models but sklearn models
        # with first -class integration for kmeans
        from sklearn.cluster import (
            KMeans,
            MiniBatchKMeans,
            DBSCAN,
            Birch,
            SpectralClustering,
            OPTICS,
            AgglomerativeClustering,
            AffinityPropagation,
            MeanShift,
            FeatureAgglomeration,
        )

        POSSIBLE_MODELS = [
            SpectralClustering,
            Birch,
            DBSCAN,
            OPTICS,
            AgglomerativeClustering,
            AffinityPropagation,
            MeanShift,
            FeatureAgglomeration,
        ]
        if is_hdbscan_available():
            import hdbscan

            if hasattr(hdbscan, "HDBSCAN"):
                POSSIBLE_MODELS.append(hdbscan.HDBSCAN)
        if model.__class__ == KMeans:

            class CentroidClusterModel(CentroidClusterBase):
                def __init__(self, model):
                    self.model: Union[KMeans, MiniBatchKMeans] = model

                def fit_predict(self, X):
                    return self.model.fit_predict(X)

                def get_centers(self):
                    return self.model.cluster_centers_

            new_model = CentroidClusterModel(model)
            return new_model

        elif model.__class__ == MiniBatchKMeans:

            class BatchCentroidClusterModel(CentroidClusterBase, BatchClusterBase):
                def __init__(self, model):
                    self.model: MiniBatchKMeans = model

                def partial_fit(self, X):
                    return self.model.partial_fit(X)

                def predict(self, X):
                    return self.model.predict(X)

                def get_centers(self):
                    return self.model.cluster_centers_

            new_model = BatchCentroidClusterModel(model)
            return new_model

        elif isinstance(model, tuple(POSSIBLE_MODELS)):
            # new_model = CentroidClusterModel(model)
            if "sklearn" in str(type(model)).lower():
                new_model = SklearnCentroidBase(model)
            elif "hdbscan" in str(type(model)).lower():
                new_model = HDBSCANClusterBase(model)
            return new_model
        elif hasattr(model, "fit_documents"):
            return model
        elif hasattr(model, "fit_predict"):
            data = {"fit_predict": model.fit_predict, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()
        elif hasattr(model, "fit_transform"):
            data = {"fit_predict": model.fit_transform, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()

    def _assign_model(self, model):
        # Check if this is a model that will fit
        # otherwise - forces a Clusterbase
        if (is_sklearn_available() or is_hdbscan_available()) and (
            "sklearn" in str(type(model)).lower()
            or "hdbscan" in str(type(model)).lower()
        ):
            model = self._assign_sklearn_model(model)
            if model is not None:
                return model

        if isinstance(model, ClusterBase):
            return model
        elif hasattr(model, "fit_documents"):
            return model
        # elif hasattr(model, "fit_predict"):
        #     # Support for SKLEARN interface
        #     data = {"fit_predict": model.fit_predict, "metadata": model.__dict__}
        #     ClusterModel = type("ClusterBase", (ClusterBase,), data)
        #     return ClusterModel()
        elif hasattr(model, "fit_predict"):
            data = {"fit_predict": model.fit_predict, "metadata": model.__dict__}
            ClusterModel = type("ClusterBase", (ClusterBase,), data)
            return ClusterModel()
        elif model is None:
            return model
        raise TypeError("Model should be inherited from ClusterBase.")

    def _token_to_auth(self, token=None):
        SIGNUP_URL = "https://cloud.relevance.ai/sdk/api"

        if os.path.exists(self._cred_fn):
            credentials = self._read_credentials()
            return credentials

        elif token:
            return self._process_token(token)

        else:
            print(f"Activation token (you can find it here: {SIGNUP_URL} )")
            if not token:
                token = getpass.getpass(f"Activation token:")
            return self._process_token(token)

    def _process_token(self, token: str):
        split_token = token.split(":")
        project = split_token[0]
        api_key = split_token[1]
        if len(split_token) > 2:
            region = split_token[3]
            base_url = self._region_to_url(region)

            if len(split_token) > 3:
                firebase_uid = split_token[4]
                return self._write_credentials(
                    project=project,
                    api_key=api_key,
                    base_url=base_url,
                    firebase_uid=firebase_uid,
                )

            else:
                return self._write_credentials(
                    project=project, api_key=api_key, base_url=base_url
                )

        else:
            return self._write_credentials(project=project, api_key=api_key)

    def _read_credentials(self):
        return json.load(open(self._cred_fn))

    def _write_credentials(self, **kwargs):
        print(
            f"Saving credentials to {self._cred_fn}. Remember to delete this file if you do not want credentials saved."
        )
        json.dump(
            kwargs,
            open(self._cred_fn, "w"),
        )
        return kwargs

    def _init_dataset(self, dataset):
        # set dataset ID and dataset attributes for consistent usage
        if isinstance(dataset, Dataset):
            self.dataset_id = dataset.dataset_id
            self.dataset: Dataset = dataset
        elif isinstance(dataset, str):
            self.dataset_id = dataset
            self.dataset = Dataset(
                project=self.project,
                api_key=self.api_key,
                dataset_id=self.dataset_id,
                firebase_uid=self.firebase_uid,
            )
        else:
            raise ValueError(
                "Dataset type needs to be either a string or Dataset instance."
            )

    @track
    def list_closest_to_center(
        self,
        dataset: Optional[Union[str, Dataset]] = None,
        vector_fields: Optional[List] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
    ):
        """
        List of documents closest from the centre.

        Parameters
        ----------

        cluster_ids: list
            Any of the cluster ids
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results

        Example
        --------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            cluster_ops.list_closest_to_center()

        """
        cluster_ids = [] if cluster_ids is None else cluster_ids
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        return self.datasets.cluster.centroids.list_closest_to_center(
            dataset_id=self._check_dataset_id(dataset),
            vector_fields=self.vector_fields
            if vector_fields is None
            else vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
        )

    @track
    def aggregate(
        self,
        vector_fields: List[str] = None,
        metrics: Optional[list] = None,
        sort: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        dataset: Optional[Union[str, Dataset]] = None,
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n

        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }

        For example, one can use the following aggregations to group score based on region and player name.

        .. code-block::

            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }

        "groupby" is the fields you want to split the data into. These are the available groupby types:

            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric

        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:

            - "avg", "max", "min", "sum", "cardinality"

        The response returned has the following in descending order. \n

        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:

            .. code-block::

                {
                    'groupby':[
                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
                        'group_size': 10, 'select_fields': ["name"]},
                    ],
                    'metrics':[
                        {'name':'Price Average','field':'price','agg':'avg'},
                    ],
                }

                # ouptut example:
                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}

        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector

        Parameters
        ----------
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            clusterer.aggregate(
                "sample_dataset_id",
                groupby=[{
                    "field": "title",
                    "agg": "wordcloud",
                }],
                vector_fields=['sample_vector_']
            )


        """
        metrics = [] if metrics is None else metrics
        sort = [] if sort is None else sort
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters

        return self.services.cluster.aggregate(
            dataset_id=self._check_dataset_id(dataset),
            vector_fields=self.vector_fields if not vector_fields else vector_fields,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            filters=filters,
            alias=self.alias,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
        )

    def list_furthest_from_center(
        self, dataset: Union[str, Dataset] = None, vector_fields: list = None
    ):
        """
        List of documents furthest from the centre.

        Parameters
        ----------
        cluster_ids: list
            Any of the cluster ids
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results

        Example
        ---------
        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            cluster_ops.list_furthest_from_center()

        """
        return self.datasets.cluster.centroids.list_furthest_from_center(
            dataset_id=self._check_dataset_id(dataset),
            vector_fields=self.vector_fields
            if vector_fields is None
            else vector_fields,
            alias=self.alias,
        )

    def _insert_centroid_documents(self):
        if hasattr(self.model, "get_centroid_documents"):
            print("Inserting centroid documents...")
            centers = self.get_centroid_documents()

            # Change centroids insertion
            results = self.services.cluster.centroids.insert(
                dataset_id=self.dataset_id,
                cluster_centers=centers,
                vector_fields=self.vector_fields,
                alias=self.alias,
            )
            self.logger.info(results)

        return

    def _check_dataset_id(self, dataset: Optional[Union[str, Dataset]] = None) -> str:
        """Helper method to get multiple dataset values"""

        if isinstance(dataset, Dataset):
            dataset_id: str = dataset.dataset_id
        elif isinstance(dataset, str):
            dataset_id = dataset
        elif dataset is None:
            if hasattr(self, "dataset_id"):
                # let's not surprise users
                print(
                    f"No dataset supplied - using last stored one '{self.dataset_id}'."
                )
                dataset_id = str(self.dataset_id)
            else:
                raise ValueError("Please supply dataset.")
        return dataset_id

    @track
    def insert_centroid_documents(
        self, centroid_documents: List[Dict], dataset: Union[str, Dataset] = None
    ):
        """
        Insert the centroid documents

        Parameters
        ------------

        centroid_documents: List[Dict]
            Insert centroid documents
        dataset: Union[str, Dataset]
            Dataset to insert

        Example
        ------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.fit_predict_update(df, vector_fields=["sample_vector_"])

            centroids = cluster_ops.get_centroid_documents()
            cluster_ops.insert_centroid_documents(centroids)

        """

        results = self.services.cluster.centroids.insert(
            dataset_id=self._check_dataset_id(dataset),
            cluster_centers=centroid_documents,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )
        return results

    @track
    def get_centroid_documents(self) -> List:
        """
        Get the centroid documents to store. This enables you to use `list_closest_to_center()`
        and `list_furthest_from_center`.

        .. code-block::

            {
                "_id": "document-id-1",
                "centroid_vector_": [0.23, 0.24, 0.23]
            }

        If multiple vector fields returns this:
        Returns multiple

        .. code-block::

            {
                "_id": "document-id-1",
                "blue_vector_": [0.12, 0.312, 0.42],
                "red_vector_": [0.23, 0.41, 0.3]
            }

        """
        if hasattr(self.model, "get_centroid_documents"):
            self.model.vector_fields = self.vector_fields
            return self.model.get_centroid_documents()
        self.centers = self.model.get_centers()

        if not hasattr(self, "vector_fields") or len(self.vector_fields) == 1:
            if isinstance(self.centers, np.ndarray):
                self.centers = self.centers.tolist()
            centroid_vector_field_name = self.vector_fields[0]
            return [
                {
                    "_id": self._label_cluster(i),
                    centroid_vector_field_name: self.centers[i],
                }
                for i in range(len(self.centers))
            ]
        # For one or more vectors, separate out the vector fields
        # centroid documents are created using multiple vector fields
        centroid_docs = []
        for i, c in enumerate(self.centers):
            centroid_doc = {"_id": self._label_cluster(i)}
            for j, vf in enumerate(self.vector_fields):
                centroid_doc[vf] = self.centers[i][vf]
            centroid_docs.append(centroid_doc.copy())
        return centroid_docs

    @property
    def centroids(self):
        """
        See your centroids if there are any.
        """
        return self.services.cluster.centroids.documents(
            self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            page_size=10000,
            # cursor: str = None,
            include_vector=True,
        )

    @track
    def delete_centroids(self, dataset: Union[str, Dataset], vector_fields: List):
        """Delete the centroids after clustering."""
        # TODO: Fix delete centroids once its moved over to Node JS
        import requests

        base_url = self.config["api.base_url"]
        response = requests.post(
            base_url + "/services/cluster/centroids/delete",
            headers={"Authorization": self.project + ":" + self.api_key},
            params={
                "dataset_id": self._check_dataset_id(dataset),
                "vector_field": vector_fields,
                "alias": self.alias,
            },
        )
        return response.json()["status"]

    def fit_predict(
        self,
        data: Union[str, Dataset, List[Dict]],
        vector_fields: List[str],
        filters: Optional[List[Dict]] = None,
        return_only_clusters: bool = True,
        include_grade: bool = False,
        update: bool = True,
        inplace: bool = False,
    ):
        """
        Parameters
        ----------
        data: Union[str, Dataset, List[Dict]]
            Either a reference to a Relevance AI Dataset, be it its name
            (string) or the object itself (Dataset), or a list of documents
            (List[Dict]).

        vector_fields: List[str]
            The vector fields over which to fit the model.

        filters: List[Dict]
            A list of filters to enable for document retrieval. This only
            applies to a reference to a Relevance AI Dataset.

        return_only_clusters: bool
            An indicator that determines what is returned. If True, this
            function returns the clusters. Else, the function returns the
            original documents.

        include_grade: bool
            An indictor that determines whether to include (True) a grade
            base on the mean silhouette score or not (False).

        update: bool
            An indicator that determines whether to update the documents
            that were part of the clustering process. This only applies to a
            reference to a Relevance AI Dataset.

        inplace: bool
            An indicator that determines whether the documents are edited
            inplace (True) or a copy is created and edited (False).

        Example
        -------

        .. code-block::

            from relevanceai import ClusterBase, Client

            client = Client()

            import random
            class CustomClusterModel(ClusterBase):
                def fit_predict(self, X):
                    cluster_labels = [random.randint(0, 100) for _ in range(len(X))]
                    return cluster_labels

            model = CustomClusterModel()

            df = client.Dataset("sample_dataset")
            clusterer = client.ClusterOps(alias="random_clustering", model=model)
            clusterer.fit_predict_update(df, vector_fields=["sample_vector_"])

        """
        filters = [] if filters is None else filters

        if update and isinstance(data, list):
            warnings.warn(
                "Cannot update list of datasets that are untethered "
                "to a Relevance AI dataset. "
                "Setting update to False."
            )
            # If data is of type List[Dict] the value of update doesn't
            # actually matter. This is more for good practice.
            update = False

        if isinstance(data, list):
            documents = data
        else:
            self._init_dataset(data)
            self.vector_fields = vector_fields
            # make sure to only get fields where vector fields exist
            filters.extend(
                [
                    {
                        "field": f,
                        "filter_type": "exists",
                        "condition": "==",
                        "condition_value": " ",
                        "strict": "must_or",
                    }
                    for f in vector_fields
                ]
            )
            # load the documents
            self.logger.warning(
                "Retrieving documents... This can take a while if the dataset is large."
            )
            print("Retrieving all documents")
            documents = self._get_all_documents(
                dataset_id=self.dataset_id, filters=filters, select_fields=vector_fields
            )
            if len(documents) == 0:
                raise NoDocumentsError()

        vectors = self._get_vectors_from_documents(vector_fields, documents)

        # Label the clusters
        print("Fitting and predicting on all relevant documents")
        cluster_labels = self._label_clusters(self.model.fit_predict(vectors))

        if include_grade:
            try:
                self._calculate_silhouette_grade(vectors, cluster_labels)
            except Exception as e:
                print(e)
                pass

        clustered_documents = self.set_cluster_labels_across_documents(
            cluster_labels,
            documents,
            inplace=inplace,
            return_only_clusters=return_only_clusters,
        )

        if not isinstance(data, list):
            if update:
                # Updating the db
                print("Updating the database...")
                results = self._update_documents(
                    self.dataset_id, clustered_documents, chunksize=10000
                )
                self.logger.info(results)

                # Update the centroid collection
                self.model.vector_fields = vector_fields

            self._insert_centroid_documents()
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

        return clustered_documents

    @track
    def fit_predict_update(
        self,
        dataset: Union[Dataset, str],
        vector_fields: List,
        filters: Optional[List] = None,
        include_grade: bool = False,
        verbose: bool = True,
    ):
        """
        This function fits a cluster model onto a dataset. It sits under `fit`
        and is usually the only function that runs unless you have a get_centroid_documents
        function.

        Parameters
        ---------------
        dataset: Union[Dataset, str],
            The dataset object to fit it on
        vector_fields: list
            The vector fields to fit the model on
        filters: list
            The filters to run it on
        include_filters_for_vector_fields: bool
            If True, only cluster on those with the vector fields inside it.

        Example
        ---------

        .. code-block::

            from relevanceai import ClusterBase, Client
            client = Client()

            import random
            class CustomClusterModel(ClusterBase):
                def fit_predict(self, X):
                    cluster_labels = [random.randint(0, 100) for _ in range(len(X))]
                    return cluster_labels

            model = CustomClusterModel()

            clusterer = client.ClusterOps(model, alias="random_clustering")
            df = client.Dataset("sample_dataset")

            clusterer.fit_predict_update(df, vector_fields=["sample_vector_"])

        """
        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        self._init_dataset(dataset)
        self.vector_fields = vector_fields

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]
        if verbose:
            print("Retrieving all documents")
        fields_to_get = vector_fields.copy()
        if self.parent_alias:
            parent_field = self._get_cluster_field_name(self.parent_alias)
            fields_to_get.append(parent_field)

        docs = self._get_all_documents(
            dataset_id=self.dataset_id, filters=filters, select_fields=fields_to_get
        )

        if len(docs) == 0:
            raise NoDocumentsError()

        if verbose:
            print("Fitting and predicting on all documents")

        clustered_docs = self.fit_predict_documents(
            vector_fields,
            docs,
            return_only_clusters=True,
            inplace=False,
            include_grade=include_grade,
        )

        # Updating the db
        if verbose:
            print("Updating the database...")
        results = self._update_documents(
            self.dataset_id, clustered_docs, chunksize=10000
        )
        self.logger.info(results)

        # Update the centroid collection
        self.model.vector_fields = vector_fields

        self._insert_centroid_documents()

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    def subcluster_predict_update(
        self,
        dataset,
        vector_fields: Optional[List] = None,
        filters: Optional[List] = None,
        verbose: bool = False,
    ):
        """

        Run subclustering on your dataset using an in-memory clustering algorithm.

        Parameters
        --------------

        dataset: Dataset
            The dataset to create
        vector_fields: List
            The list of vector fields to run fitting, prediction and updating on
        filters: Optional[List]
            The list of filters to run clustering on
        verbose: bool
            If True, this should be verbose

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")

            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=10)
            clusterer = ClusterOps(alias="minibatchkmeans-10", model=model)
            clusterer.subcluster_predict_update(
                dataset=ds,
            )

        """
        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        self._init_dataset(dataset)
        self.vector_fields = vector_fields  # type: ignore

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields  # type: ignore
        ]

        if verbose:
            print("Fitting and predicting on all documents")
        # Here we run subfitting on these documents
        clustered_docs = self.subcluster_predict_documents(
            vector_fields=vector_fields, filters=filters, verbose=verbose
        )

        # Updating the db
        # print("Updating the database...")
        # results = self._update_documents(
        #     self.dataset_id, clustered_docs, chunksize=10000
        # )
        # self.logger.info(results)

        # # Update the centroid collection
        # self.model.vector_fields = vector_fields

        # self._insert_centroid_documents()

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    def subpartialfit_predict_update(
        self,
        dataset,
        vector_fields: list,
        filters: Optional[list] = None,
        cluster_ids: Optional[list] = None,
        verbose: bool = True,
    ):
        """

        Run partial fit subclustering on your dataset.

        Parameters
        ------------

        dataset: Dataset
            The dataset to call fit predict update on
        vector_fields: list
            The list of vector fields
        filters: list
            The list of filters

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()

            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")
            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=10)
            clusterer = ClusterOps(alias="minibatchkmeans-10", model=model)
            clusterer.subpartialfit_predict_update(
                dataset=ds,
            )

        """
        # Get data

        self._init_dataset(dataset)

        filters = [] if filters is None else filters
        cluster_ids = [] if cluster_ids is None else cluster_ids
        # Loop through each unique cluster ID and run clustering
        parent_field = self._get_cluster_field_name(self.parent_alias)

        print("Getting unique cluster IDs...")
        unique_clusters = self.unique_cluster_ids(alias=self.parent_alias)

        for i, unique_cluster in enumerate(tqdm(unique_clusters)):
            cluster_filters = filters.copy()
            cluster_filters += [
                {
                    "field": parent_field,
                    "filter_type": "category",
                    "condition": "==",
                    "condition_value": unique_cluster,
                }
            ]
            self.partial_fit_predict_update(
                dataset=self.dataset_id,
                vector_fields=vector_fields,
                filters=cluster_filters,
                verbose=False,
            )

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    def subcluster_predict_documents(
        self,
        vector_fields: Optional[List] = None,
        filters: Optional[List] = None,
        cluster_ids: Optional[List] = None,
        verbose: bool = True,
    ):
        """
        Subclustering using fit predict update. This will loop through all of the
        different clusters and then run subclustering on them. For this, you need to

        Parameters
        ------------

        Example
        ---------

        ..code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")

            # Creating 100 sample documents
            from relevanceai.package_utils.datasets import mock_documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            # Run simple clustering first
            ds.auto_cluster("kmeans-3", vector_fields=["sample_1_vector_"])

            # Start KMeans
            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=20)

            # Run subclustering.
            cluster_ops = client.ClusterOps(
                alias="subclusteringkmeans",
                model=model,
                parent_alias="kmeans-3")


        """
        filters = [] if filters is None else filters
        cluster_ids = [] if cluster_ids is None else cluster_ids
        # Loop through each unique cluster ID and run clustering
        parent_field = self._get_cluster_field_name(self.parent_alias)

        if verbose:
            print("Getting unique cluster IDs...")
        if not cluster_ids:
            unique_clusters = self.unique_cluster_ids(alias=self.parent_alias)
        else:
            unique_clusters = cluster_ids

        for i, unique_cluster in enumerate(tqdm(unique_clusters)):
            cluster_filters = filters.copy()
            cluster_filters += [
                {
                    "field": parent_field,
                    "filter_type": "category",
                    "condition": "==",
                    "condition_value": unique_cluster,
                }
            ]
            self.fit_predict_update(
                dataset=self.dataset_id,
                vector_fields=vector_fields,
                filters=cluster_filters,
                include_grade=False,
                verbose=False,
            )

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    def _check_for_dataset_id(self):
        if not hasattr(self, "dataset_id"):
            raise ValueError(
                "You are missing a dataset ID. Please set using the argument dataset_id='...'."
            )

    def unique_cluster_ids(
        self,
        alias: str = None,
        minimum_cluster_size: int = 3,
        dataset_id: str = None,
        num_clusters: int = 1000,
    ):
        """
        We call facets on our data, which looks a little like this:

        {'results': {'_cluster_.sample_1_vector_.kmeans-3': [{'_cluster_.sample_1_vector_.kmeans-3': 'cluster-2',
                'frequency': 41,
                'value': 'cluster-2'},
            {'_cluster_.sample_1_vector_.kmeans-3': 'cluster-0',
                'frequency': 34,
                'value': 'cluster-0'},
            {'_cluster_.sample_1_vector_.kmeans-3': 'cluster-1',
                'frequency': 25,
                'value': 'cluster-1'}]}}
        """
        # Mainly to be used for subclustering
        # Get the cluster alias
        if dataset_id is None:
            self._check_for_dataset_id()
            dataset_id = self.dataset_id

        cluster_field = self._get_cluster_field_name(alias=alias)

        # currently the logic for facets is that when it runs out of pages
        # it just loops - therefore we need to store it in a simple hash
        # and then add them to a list
        all_cluster_ids: Set = set()

        while len(all_cluster_ids) < num_clusters:
            facet_results = self.datasets.facets(
                dataset_id=dataset_id,
                fields=[cluster_field],
                page_size=int(self.config["data.max_clusters"]),
                page=1,
                asc=True,
            )
            if "results" in facet_results:
                facet_results = facet_results["results"]
            if cluster_field not in facet_results:
                raise ValueError(
                    f"No clusters with alias `{alias}`. Please check the schema."
                )
            for facet in facet_results[cluster_field]:
                if facet["frequency"] > minimum_cluster_size:
                    curr_len = len(all_cluster_ids)
                    all_cluster_ids.add(facet[cluster_field])
                    new_len = len(all_cluster_ids)
                    if new_len == curr_len:
                        return list(all_cluster_ids)
        return list(all_cluster_ids)

    @track
    def fit_dataset(
        self,
        dataset,
        vector_fields,
        filters: Optional[list] = None,
        return_only_clusters: bool = True,
        inplace=False,
        verbose: bool = True,
    ):
        """ """
        filters = [] if filters is None else filters

        # load the documents
        self.logger.warning(
            "Retrieving documents... This can take a while if the dataset is large."
        )

        self._init_dataset(dataset)
        self.vector_fields = vector_fields

        # make sure to only get fields where vector fields exist
        filters += [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]
        if verbose:
            print("Retrieving all documents")
        docs = self._get_all_documents(
            dataset_id=self.dataset_id, filters=filters, select_fields=vector_fields
        )

        if verbose:
            print("Fitting and predicting on all documents")
        return self.fit_predict_documents(
            vector_fields,
            docs,
            return_only_clusters=True,
            inplace=inplace,
        )

    def _concat_vectors_from_list(self, list_of_vectors: list):
        """Concatenate 2 vectors together in a pairwise fashion"""
        return [np.concatenate(x) for x in list_of_vectors]

    def _get_vectors_from_documents(self, vector_fields: list, documents: List[Dict]):
        if len(vector_fields) == 1:
            # filtering out entries not containing the specified vector
            documents = list(filter(DocUtils.list_doc_fields, documents))
            vectors = self.get_field_across_documents(
                vector_fields[0], documents, missing_treatment="skip"
            )
        else:
            # In multifield clusering, we get all the vectors in each document
            # (skip if they are missing any of the vectors)
            # Then run clustering on the result
            documents = list(self.filter_docs_for_fields(vector_fields, documents))
            all_vectors = self.get_fields_across_documents(
                vector_fields, documents, missing_treatment="skip_if_any_missing"
            )
            # Store the vector field lengths to de-concatenate them later
            self._vector_field_length: dict = {}
            prev_vf = 0
            for i, vf in enumerate(self.vector_fields):
                self._vector_field_length[vf] = {}
                self._vector_field_length[vf]["start"] = prev_vf
                end_vf = prev_vf + len(all_vectors[0][i])
                self._vector_field_length[vf]["end"] = end_vf
                # Update the ending
                prev_vf = end_vf

            # Store the vector lengths
            vectors = self._concat_vectors_from_list(all_vectors)

        return vectors

    @track
    def partial_fit_documents(
        self,
        vector_fields: list,
        documents: List[Dict],
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.

        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        docs: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document
        inplace: bool
            If True, the documents are edited inplace otherwise, a copy is made first
        kwargs: dict
            Any other keyword argument will go directly into the clustering algorithm

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="batchkmeans_2", model=model)

            cluster_ops.parital_fit(df, vector_fields=["documentation_vector_"])
            cluster_ops.predict_update(df, vector_fields=["sample_vector_"])

        """
        self.vector_fields = vector_fields

        vectors = self._get_vectors_from_documents(vector_fields, documents)

        self.model.partial_fit(vectors)

    def _chunk_dataset(
        self,
        dataset: Dataset,
        select_fields: Optional[list] = None,
        chunksize: int = 100,
        filters: Optional[list] = None,
    ):
        """Utility function for chunking a dataset"""
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        cursor = None

        docs = self._get_documents(
            dataset_id=self.dataset_id,
            include_cursor=True,
            number_of_documents=chunksize,
            select_fields=select_fields,
            filters=filters,
        )

        while len(docs["documents"]) > 0:
            yield docs["documents"]
            docs = self._get_documents(
                dataset_id=self.dataset_id,
                cursor=docs["cursor"],
                include_cursor=True,
                select_fields=select_fields,
                number_of_documents=chunksize,
                filters=filters,
            )

    @track
    def partial_fit_dataset(
        self,
        dataset: Union[str, Dataset],
        vector_fields: List[str],
        chunksize: int = 100,
        filters: Optional[list] = None,
    ):
        """
        Fit The dataset by partial documents.


        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="minibatchkmeans_2", model=model)

            cluster_ops.partial_fit_dataset(df, vector_fields=["documentation_vector_"])

        """
        filters = [] if filters is None else filters

        self.vector_fields = vector_fields
        if len(vector_fields) > 1:
            raise ValueError(
                "We currently do not support multiple vector fields on partial fit"
            )

        if isinstance(dataset, str):
            self.dataset = Dataset(
                project=self.project,
                api_key=self.api_key,
                dataset_id=dataset,
                firebase_uid=self.firebase_uid,
            )
        else:
            self.dataset = dataset

        filters = [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ] + filters

        for c in self._chunk_dataset(
            self.dataset, self.vector_fields, chunksize=chunksize, filters=filters
        ):
            vectors = self._get_vectors_from_documents(vector_fields, c)
            self.model.partial_fit(vectors)

    @track
    def partial_fit_predict_update(
        self,
        dataset: Union[Dataset, str],
        vector_fields: Optional[List[str]] = None,
        chunksize: int = 100,
        filters: Optional[List] = None,
        verbose: bool = True,
    ):
        """
        Fit, predict and update on a dataset.
        Users can also start to run these separately one by one.

        Parameters
        --------------

        dataset: Union[Dataset]
            The dataset class

        vector_fields: List[str]
            The list of vector fields

        chunksize: int
            The size of the chunks

        Example
        -----------

        .. code-block::

            # Real-life example from Research Dashboard
            from relevanceai import Client
            client = Client()
            df = client.Dataset("research2vec")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=50)
            cluster_ops = client.ClusterOps(alias="minibatchkmeans_50", model=model)

            cluster_ops.partial_fit_predict_update(
                df,
                vector_fields=['title_trainedresearchqgen_vector_'],
                chunksize=1000
            )

        """
        vector_fields = [] if vector_fields is None else vector_fields
        filters = [] if filters is None else filters

        if verbose:
            print("Fitting dataset...")
        self.partial_fit_dataset(
            dataset=dataset,
            vector_fields=vector_fields,
            chunksize=chunksize,
            filters=filters,
        )
        if verbose:
            print("Updating your dataset...")
        self.predict_update(dataset=dataset)
        if hasattr(self.model, "get_centers"):
            if verbose:
                print("Inserting your centroids...")
            self.insert_centroid_documents(
                self.get_centroid_documents(), dataset=dataset
            )

        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )

    @track
    def predict_documents(
        self,
        vector_fields,
        documents,
        return_only_clusters: bool = True,
        inplace: bool = True,
    ):
        """
        Predict the documents
        """
        if len(vector_fields) > 1:
            raise ValueError("Currently do not suport more than 1 vector field.")
        vectors = self._get_vectors_from_documents(vector_fields, documents)
        cluster_labels = self.model.predict(vectors)
        cluster_labels = self._label_clusters(cluster_labels)
        return self.set_cluster_labels_across_documents(
            cluster_labels,
            documents,
            inplace=inplace,
            return_only_clusters=return_only_clusters,
        )

    @track
    def predict_update(
        self,
        dataset,
        vector_fields: Optional[List[str]] = None,
        chunksize: int = 20,
        verbose: bool = True,
    ):
        """
        Predict the dataset.

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="minibatchkmeans_2", model=model)

            cluster_ops.partial_fit_dataset(df)
            cluster_ops.predict_dataset(df)

        """
        if not vector_fields:
            vector_fields = self.vector_fields

        all_responses = {
            "inserted": 0,
            "failed_documents": [],
            "failed_documents_detailed": [],
        }

        filters = [
            {
                "field": f,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for f in vector_fields
        ]

        for c in self._chunk_dataset(
            dataset, self.vector_fields, chunksize=chunksize, filters=filters
        ):
            cluster_predictions = self.predict_documents(
                vector_fields=vector_fields, documents=c
            )
            response = self.dataset._update_documents(
                dataset_id=self.dataset_id, documents=cluster_predictions
            )
            for k, v in response.items():
                if isinstance(all_responses[k], int):
                    all_responses["inserted"] += v
                elif isinstance(all_responses[k], list):
                    all_responses[k] += v
        if verbose:
            print(
                "Build your clustering app here: "
                + f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
            )
        return all_responses

    @track
    def fit_predict_documents(
        self,
        vector_fields: list,
        documents: List[Dict],
        return_only_clusters: bool = True,
        inplace: bool = True,
        include_grade: bool = False,
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.

        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        docs: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document
        inplace: bool
            If True, the documents are edited inplace otherwise, a copy is made first
        kwargs: dict
            Any other keyword argument will go directly into the clustering algorithm

        Example
        -----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import MiniBatchKMeans
            model = MiniBatchKMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="minibatchkmeans_2", model=model)

            cluster_ops.fit_predict_documents(df, vector_fields=["documentation_vector_"])

        """
        self.vector_fields = vector_fields

        vectors = self._get_vectors_from_documents(vector_fields, documents)

        cluster_labels = self.model.fit_predict(vectors)

        if not self.parent_alias:
            cluster_labels_values = self._label_clusters(cluster_labels)
        else:
            prev_cluster_labels = self._get_parent_cluster_values(
                vector_fields=vector_fields,
                alias=self.parent_alias,
                documents=documents,
            )
            cluster_labels_values = self._label_subclusters(
                labels=cluster_labels, prev_cluster_labels=prev_cluster_labels
            )

        if include_grade:
            try:
                self._calculate_silhouette_grade(vectors, cluster_labels_values)
            except Exception as e:
                print(e)
                pass
        return self.set_cluster_labels_across_documents(
            cluster_labels=cluster_labels_values,
            documents=documents,
            inplace=inplace,
            return_only_clusters=return_only_clusters,
        )

    def _get_parent_cluster_values(
        self, vector_fields: list, alias: str, documents
    ) -> list:
        field = ".".join([self.cluster_field, ".".join(sorted(vector_fields)), alias])
        return self.get_field_across_documents(
            field, documents, missing_treatment="skip"
        )

    @staticmethod
    def _calculate_silhouette_grade(vectors, cluster_labels):
        from relevanceai.reports.cluster_report.grading import get_silhouette_grade
        from sklearn.metrics import silhouette_samples

        score = silhouette_samples(vectors, cluster_labels, metric="euclidean").mean()
        grade = get_silhouette_grade(score)

        print("---------------------------")
        print(f"Grade: {grade}")
        print(f"Mean Silhouette Score: {score}")
        print("---------------------------")

    def set_cluster_labels_across_documents(
        self,
        cluster_labels: list,
        documents: List[Dict],
        inplace: bool = True,
        return_only_clusters: bool = True,
    ):
        """
        Utility function to allow users to set cluster labels

        Parameters
        ------------
        cluster_labels: List[str, int]
            A list of integers of string. If it is an integer - it will automatically add a 'cluster-' prefix
            to help avoid incorrect data type parsing. You can override this behavior by setting clusters
            as strings.
        documents: List[dict]
            When the documents are in
        inplace: bool
            If True, then the clusters are set in place.
        return_only_clusters: bool
            If True, then the return_only_clusters will return documents with just the cluster field and ID.
            This can be helpful when you want to upsert quickly without having to re-insert the entire document.

        Example
        -----------

        .. code-block::

            labels = list(range(10))
            documents = [{"_id": str(x)} for x in range(10)]
            cluster_ops.set_cluster_labels_across_documents(labels, documents)

        """
        if inplace:
            self._set_cluster_labels_across_documents(cluster_labels, documents)
            if return_only_clusters:
                return [
                    {"_id": d.get("_id"), self.cluster_field: d.get(self.cluster_field)}
                    for d in documents
                ]
            return documents

        # useful if you want to upload as quickly as possible
        new_documents = documents.copy()

        self._set_cluster_labels_across_documents(cluster_labels, new_documents)
        if return_only_clusters:
            return [
                {"_id": d.get("_id"), self.cluster_field: d.get(self.cluster_field)}
                for d in new_documents
            ]
        return new_documents

    def _get_cluster_field_name(self, alias: str = None):
        if alias is None:
            alias = self.alias
        if isinstance(self.vector_fields, list):
            set_cluster_field = (
                f"{self.cluster_field}.{'.'.join(self.vector_fields)}.{alias}"
            )
        elif isinstance(self.vector_fields, str):
            set_cluster_field = f"{self.cluster_field}.{self.vector_fields}.{alias}"
        return set_cluster_field

    def _set_cluster_labels_across_documents(self, cluster_labels, documents):
        set_cluster_field = self._get_cluster_field_name()
        self.set_field_across_documents(set_cluster_field, cluster_labels, documents)

    def _label_cluster(self, label: Union[int, str]):
        if not isinstance(label, str):
            return "cluster-" + str(label)
        return str(label)

    def _label_subcluster(self, label: Union[int, str], prev_cluster_label) -> str:
        return prev_cluster_label + "-" + str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]

    def _label_subclusters(self, labels: List[str], prev_cluster_labels: List[str]):
        return [
            self._label_subcluster(label, prev_cluster_label)
            for label, prev_cluster_label in zip(labels, prev_cluster_labels)
        ]

    @property
    def metadata(self):
        """
        If metadata is none, retrieves metadata about a dataset. notably description, data source, etc
        Otherwise, you can store the metadata about your cluster here.

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("_github_repo_vectorai")
            from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel

            model = KMeansModel()
            kmeans = client.ClusterOps(model, alias="kmeans_sample")
            kmeans.fit(df, vector_fields=["sample_1_vector_"])
            kmeans.metadata
            # {"k": 10}

        """
        return self.services.cluster.centroids.metadata(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
        )

    @metadata.setter
    def metadata(self, metadata: dict):
        """
        If metadata is none, retrieves metadata about a dataset. notably description, data source, etc
        Otherwise, you can store the metadata about your cluster here.

        Parameters
        ----------
        metadata: dict
           If None, it will retrieve the metadata, otherwise
           it will overwrite the metadata of the cluster

        Example
        ----------

        .. code-block::


            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)

            cluster_ops.fit(df, vector_fields=["sample_1_vector_"])
            cluster_ops.metadata
            # {"k": 10}

        """
        return self.services.cluster.centroids.metadata(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            metadata=metadata,
        )

    @track
    def evaluate(
        self,
        ground_truth_column: Union[str, None] = None,
        metric: Union[str, Callable] = "euclidean",
        random_state: Union[int, None] = None,
        average_score: bool = False,
    ):
        """
        Evaluate your clusters using the silhouette score, or if true labels are provided, using completeness, randomness and homogeniety as well

        Parameters
        ----------
        ground_truth_index : str
            the index of the true label of each sample in dataset

        metric : str or Callable
            A string referencing suportted distance functions, or custom method for calculating the distance between 2 vectors

        random_state : int, default None
            for reproducability

        average_score : bool
            a boolean that determines whether to average the evaluation metrics in a new a score. only applicable if ground_truth_index is not None

        Example
        ----------

        .. code-block::


            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset")

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)

            cluster_ops.fit(df, vector_fields=["sample_vector_"])

            cluster_ops.evaluate()
            cluster_ops.evaluate("truth_column")
        """

        from sklearn.metrics import (
            adjusted_rand_score,
            completeness_score,
            silhouette_score,
            homogeneity_score,
        )

        vector_field = self.vector_fields[0]
        cluster_field = self.cluster_field
        alias = self.alias

        samples = self._get_all_documents(self.dataset_id, include_vector=True)

        vectors = [sample[vector_field] for sample in samples]
        pred_labels = [sample[cluster_field][vector_field][alias] for sample in samples]

        def norm(value, min, max):
            return (value - min) / (max - min)

        s_score = silhouette_score(
            vectors, pred_labels, metric=metric, random_state=random_state
        )

        stats = {}
        stats["silhouette"] = {
            "score": s_score,
            "description": METRIC_DESCRIPTION["silhouette"],
        }

        if (
            ground_truth_column
        ):  # if the ground truth column was provided, we can run these cluster evaluation measures

            ground_truth_labels = {  # convert cluster1 etc. to respective label found in ground truth column
                key: value[0][ground_truth_column]
                for key, value in self.list_closest_to_center(page_size=1).items()
                if value
            }
            pred_labels = (
                [  # get the predicted clusters and return their respective true labels
                    ground_truth_labels[sample[cluster_field][vector_field][alias]]
                    for sample in samples
                ]
            )
            true_labels = [
                sample[ground_truth_column] for sample in samples
            ]  # same for actual labels

            ar_score = adjusted_rand_score(true_labels, pred_labels)  # compute metric
            c_score = completeness_score(true_labels, pred_labels)  # compute metric
            h_score = homogeneity_score(true_labels, pred_labels)  # compute metric

            stats["random"] = {
                "score": ar_score,
                "description": METRIC_DESCRIPTION["random"],
            }
            stats["completeness"] = {
                "score": c_score,
                "description": METRIC_DESCRIPTION["completeness"],
            }
            stats["homogeneity"] = {
                "score": h_score,
                "description": METRIC_DESCRIPTION["homogeneity"],
            }

            if average_score:  # If the user wants an average as well
                ns_score = norm(s_score, min=-1, max=1)
                nar_score = norm(ar_score, min=-1, max=1)

                average_score = (ns_score + nar_score + c_score + h_score) / 4

                stats["average"] = {
                    "score": average_score,
                    "description": METRIC_DESCRIPTION["average"],
                }

        return stats

    @beta
    @track
    def store_cluster_report(
        self, report_name: str, report: dict, verbose: bool = True
    ):
        """

        Store the cluster data.

        .. code-block::

            from relevanceai import Client
            client = Client()
            client.store_cluster_report("sample", {"value": 3})

        """
        response: dict = self.reports.clusters.create(
            name=report_name, report=self.json_encoder(report)
        )
        if verbose:
            print(
                f"You can now access your report at https://cloud.relevance.ai/report/cluster/{self.region}/{response['_id']}"
            )
        return response

    @track
    def internal_report(self, verbose: bool = True):
        """
        Get a report on your clusters.

        Example
        ---------

        This is what is returned on an `auto_cluster` method.

        .. code-block::

            from relevanceai.package_utils.datasets import mock_documents
            docs = mock_documents(10)
            df = client.Dataset('sample')
            df.upsert_documents(docs)
            cluster_ops = df.auto_cluster('kmeans-2', ['sample_1_vector_'])
            cluster_ops.internal_report()

        This is what is returned on an `internal_report` method.

        .. code-block::

            from relevanceai import Client
            # client = Client()
            from relevanceai.package_utils.datasets import mock_documents
            ds = client.Dataset("sample")
            # Creates 100 sample documents
            documents = mock_documents(100)
            ds.upsert_documents(documents)

            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=10)
            clusterer = client.ClusterOps(alias="not-auto-kmeans-10", model=model)

            clusterer.fit_predict_update(dataset=ds, vector_fields=["sample_1_vector_"])
            cluster_ops.internal_report()

        """
        if isinstance(self.vector_fields, list) and len(self.vector_fields) > 1:
            raise ValueError(
                "We currently do not support more than 1 vector field when reporting."
            )
        from relevanceai.reports.cluster_report import ClusterReport

        # X is all the vectors
        cluster_field_name = self._get_cluster_field_name()
        all_docs = self._get_all_documents(
            self.dataset_id, select_fields=self.vector_fields + [cluster_field_name]
        )
        cluster_labels = self.get_field_across_documents(cluster_field_name, all_docs)
        self.number_of_clusters = len(set(cluster_labels))
        self._report = ClusterReport(
            self.get_field_across_documents(self.vector_fields[0], all_docs),
            cluster_labels=self.get_field_across_documents(
                cluster_field_name, all_docs
            ),
            model=self.model,
            num_clusters=self.number_of_clusters,
        )
        cluster_response = self.store_cluster_report(
            report_name=cluster_field_name,
            report=self.json_encoder(self._report.internal_report),
            verbose=verbose,
        )

        return self._report.internal_report

    report = internal_report

    def operate(self, field: str, func: Callable, output_field: Optional[str] = None):
        """
        Run an function per cluster.

        Parameters
        ------------

        field: str
            The field to operate on
        func: Callable
            The function to run on all the values once they are received.
            It should take in a list of values
        output_field: Optional[str]
            Outputs for every document in the cluster

        Example
        ---------

        .. code-block::

            import numpy as np
            def vector_mean(vectors):
                return np.mean(vectors, axis=0)
            cluster_centroids = cluster_ops.operate(
                field="review_sentence_answer_use_vector_",
                func=vector_mean
            )

        """
        # Run a function on each cluster
        output: Dict = {}
        cluster_ids = self.unique_cluster_ids()
        for cluster_id in tqdm(cluster_ids):
            self._operate(cluster_id, field, output, func)
        if output_field is not None:
            self.update_documents_within_clusters(output, output_field)
        return output

    def _operate(self, cluster_id: str, field: str, output: dict, func: Callable):
        """
        Internal function for operations
        """
        cluster_field = self._get_cluster_field_name()
        # TODO; change this to fetch all documents
        documents = self.datasets.documents.get_where(
            self.dataset_id,
            filters=[
                {
                    "field": cluster_field,
                    "filter_type": "exact_match",
                    "condition": "==",
                    "condition_value": cluster_id,
                },
                {
                    "field": field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
            select_fields=[field, cluster_field],
            page_size=9999,
        )
        # get the field across each
        arr = self.get_field_across_documents(field, documents["documents"])
        output[cluster_id] = func(arr)

    def create_centroids(
        self, vector_fields: List[str], operation: Optional[Callable] = None
    ):
        """
        Create centroids if there are none. The default operation is to take the centroid
        of each vector.
        An alternative function can be provided provided.

        Example of the operation in question for mean:

        .. code-block::

            def vector_mean(vectors):
                return np.mean(vectors, axis=0)

            clusterops.create_centroids(["sample_vector_"], operation=vector_mean)

        """
        if len(vector_fields) > 1:
            raise ValueError("currently do not support more than 1 vector field.")
        vector_field: str = vector_fields[0]

        def vector_mean(vectors):
            return np.mean(vectors, axis=0)

        if operation == "mean":
            operation = vector_mean

        cluster_centroids = self.operate(field=vector_field, func=operation)  # type: ignore
        centroid_docs = []
        for k, v in cluster_centroids.items():
            centroid_docs.append({"_id": str(k), vector_field: v.tolist()})
        return self.insert_centroid_documents(centroid_docs)

    def _get_filter_for_cluster(self, cluster_id):
        cluster_field = self._get_cluster_field_name()
        filters = [
            {
                "field": cluster_field,
                "filter_type": "exact_match",
                "condition": "==",
                "condition_value": cluster_id,
            }
        ]
        return filters

    def update_documents_within_cluster(self, cluster_id: str, update: dict):
        """
        Update all the documents within a cluster
        """
        cluster_filter = self._get_filter_for_cluster(cluster_id)
        result = self.datasets.documents.update_where(
            dataset_id=self.dataset_id, update=update, filters=cluster_filter
        )
        return result

    def update_documents_within_clusters(
        self, cluster_id_and_updates: dict, output_field: str = None
    ):
        """
        Takes the cluster ids and updates and updates them accordingly

        Example
        ---------

        .. code-block::

            # Let us take the operator results
            operator_results = clusterops.operate(...)
            clusterops.update_documents_within_clusters(
                operator_results, output_field="centroid_vector_"
            )
        """
        results = []
        for cluster_id, update_value in cluster_id_and_updates.items():
            if type(update_value) != dict:
                if output_field is None:
                    raise ValueError(
                        """
                        As update value is not a dictionary, you will need to
                        specify `output_field=`.
                    """
                    )
                update: dict = {}
                self.set_field(output_field, update, update_value)
            else:
                update = update_value

            cluster_filter = self._get_filter_for_cluster(cluster_id)
            result = self.datasets.documents.update_where(
                dataset_id=self.dataset_id, update=update, filters=cluster_filter
            )
            results.append(result)
        return results



"""Utilities around packages
"""
from relevanceai.package_utils.analytics_funcs import *
from relevanceai.package_utils.base import *
from relevanceai.package_utils.cache import *
from relevanceai.package_utils.concurrency import *
from relevanceai.package_utils.config import *
from relevanceai.package_utils.constants import *
from relevanceai.package_utils.datasets import *
from relevanceai.package_utils.errors import *
from relevanceai.package_utils.fire_and_forget import *
from relevanceai.package_utils.integration_checks import *
from relevanceai.package_utils.json_encoder import *
from relevanceai.package_utils.logger import *
from relevanceai.package_utils.make_id import *
from relevanceai.package_utils.print_formats import *
from relevanceai.package_utils.progress_bar import *
from relevanceai.package_utils.transport import *
from relevanceai.package_utils.utils import *
from relevanceai.package_utils.version_decorators import *
from relevanceai.package_utils.warnings import *



"""
Test the clustering workflow from getting the documents, clustering and then inserting the relevant centroids
"""

import pytest

from relevanceai import Client
from relevanceai.interfaces import Dataset

from typing import List

from relevanceai.workflows.cluster_ops.clusterops import ClusterOps
from relevanceai.workflows.cluster_ops.cluster_base import CentroidClusterBase


@pytest.mark.skip(reason="no longer supported")
@pytest.mark.parametrize(
    "vector_fields", [["sample_1_vector_"], ["sample_2_vector_", "sample_1_vector_"]]
)
def test_cluster_integration_one_liner(
    test_client: Client, vector_dataset_id: str, vector_fields: List
):
    """Smoke Test for the entire clustering workflow."""
    # Retrieve a previous dataset
    VECTOR_FIELDS = vector_fields
    test_client.vector_tools.cluster.kmeans_cluster(
        dataset_id=vector_dataset_id,
        vector_fields=VECTOR_FIELDS,
        overwrite=True,
        alias="sample_cluster",
    )
    assert True



import pytest

from relevanceai.interfaces import Dataset

from .utils import VECTOR_FIELDS

ORIGINAL_ALIAS = "minibatchkmeans-3"


@pytest.fixture(scope="function")
def minibatch_clusterer(test_df: Dataset):
    clusterer = test_df.auto_cluster(ORIGINAL_ALIAS, vector_fields=VECTOR_FIELDS)
    yield clusterer
    clusterer.delete_centroids(test_df.dataset_id, VECTOR_FIELDS)


@pytest.fixture(scope="function")
def minibatch_subclusterer(test_df: Dataset):
    # Running batch k means after clustering
    ALIAS = "minibatchkmeans-4"
    clusterer = test_df.auto_cluster(ORIGINAL_ALIAS, vector_fields=VECTOR_FIELDS)
    clusterer = test_df.auto_cluster(
        ALIAS,
        vector_fields=VECTOR_FIELDS,
        parent_alias=ORIGINAL_ALIAS,
    )
    yield test_df, ALIAS
    clusterer.delete_centroids(test_df.dataset_id, VECTOR_FIELDS)


@pytest.fixture(scope="function")
def kmeans_subclusterer(test_df: Dataset):
    # Running K Means after clustering
    ALIAS = "kmeans-4"
    clusterer = test_df.auto_cluster(ORIGINAL_ALIAS, vector_fields=VECTOR_FIELDS)
    clusterer = test_df.auto_cluster(
        ALIAS,
        vector_fields=VECTOR_FIELDS,
        parent_alias=ORIGINAL_ALIAS,
    )
    yield test_df, ALIAS
    # clusterer.delete_centroids(test_df.dataset_id, VECTOR_FIELDS)



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
from relevanceai.dataset.auto.cluster import Cluster
from relevanceai.dataset.auto.labels import Labels
from relevanceai.dataset.auto.reduce_dimensions import DimensionalityReduction
from relevanceai.dataset.auto.vectorize import Vectorize
from relevanceai.dataset.auto.community_detection import CommunityDetection
from relevanceai.dataset.auto.scaling import Scale


class Operations(
    Cluster, Labels, DimensionalityReduction, Vectorize, CommunityDetection, Scale
):
    pass



import analytics
import asyncio
import json
import os
import copy

from typing import Callable
from base64 import b64decode as decode
from functools import wraps

from relevanceai.package_utils.config import CONFIG
from relevanceai.package_utils.json_encoder import json_encoder
from relevanceai.package_utils.fire_and_forget import fire_and_forget


def is_tracking_enabled():
    if CONFIG.is_field("mixpanel.is_tracking_enabled", CONFIG.config):
        return CONFIG.get_field("mixpanel.is_tracking_enabled", CONFIG.config)


def get_json_size(json_obj):
    # Returns it in bytes
    return len(json.dumps(json_obj).encode("utf-8")) / 1024


TRANSIT_ENV_VAR = "_IS_ANALYTICS_IN_TRANSIT"


def track(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if os.getenv(TRANSIT_ENV_VAR) == "TRUE":
            return func(*args, **kwargs)

        os.environ[TRANSIT_ENV_VAR] = "TRUE"

        try:
            if is_tracking_enabled():

                @fire_and_forget
                def send_analytics():
                    properties = {}
                    if "firebase_uid" in kwargs:
                        user_id = kwargs["firebase_uid"]
                    elif hasattr(args[0], "firebase_uid"):
                        user_id = args[0].firebase_uid
                    else:
                        user_id = "firebase_uid_not_detected"

                    event_name = f"pysdk-{func.__name__}"

                    # kwargs.update(dict(zip(func.__code__.co_varnames, args)))
                    # all_kwargs = copy.deepcopy(kwargs)
                    # all_kwargs = all_kwargs.update(
                    #     dict(zip(func.__code__.co_varnames, args))
                    # )

                    # if hasattr(func, "__code__"):
                    additional_args = dict(zip(func.__code__.co_varnames, args))

                    if "dataset_id" in kwargs:
                        properties["dataset_id"] = kwargs["dataset_id"]

                    elif "dataset_id" in additional_args:
                        properties["dataset_id"] = additional_args["dataset_id"]

                    elif hasattr(args[0], "dataset_id"):
                        properties["dataset_id"] = args[0].dataset_id

                    elif "self" in additional_args:
                        if hasattr(additional_args["self"], "dataset_id"):
                            properties["dataset_id"] = additional_args[
                                "self"
                            ].dataset_id

                    full_properties = copy.deepcopy(properties)
                    full_properties.update(
                        {
                            "additional_args": additional_args,
                            "args": args,
                            "kwargs": kwargs,
                        }
                    )
                    if user_id is not None:
                        # TODO: Loop through the properties and remove anything
                        # greater than 5kb
                        if (
                            "insert" in event_name
                            or "upsert" in event_name
                            or "update" in event_name
                            or "fit" in event_name
                            or "predict" in event_name
                            or get_json_size(
                                json_encoder(full_properties, force_string=True)
                            )
                            > 30
                        ):
                            response = analytics.track(
                                user_id=user_id,
                                event=event_name,
                                properties=json_encoder(properties, force_string=True),
                            )
                        else:
                            response = analytics.track(
                                user_id=user_id,
                                event=event_name,
                                properties=json_encoder(
                                    full_properties, force_string=True
                                ),
                            )

                send_analytics()
                # asyncio.ensure_future(send_analytics())
        except Exception as e:
            pass
        try:
            return func(*args, **kwargs)
        finally:
            os.environ[TRANSIT_ENV_VAR] = "FALSE"

    return wrapper


def track_event_usage(event_name: str):
    EVENT_NAME = event_name
    write_key = CONFIG.get_field("mixpanel.write_key", CONFIG.config)
    analytics.write_key = decode(write_key).decode("utf-8")

    def track(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if os.getenv(TRANSIT_ENV_VAR) == "TRUE":
                return func(*args, **kwargs)
            os.environ[TRANSIT_ENV_VAR] = "TRUE"
            try:
                if is_tracking_enabled():

                    async def send_analytics():
                        user_id = "OPEN_SOURCE_USER"
                        event_name = f"pysdk-{EVENT_NAME}"

                        if user_id is not None:
                            # TODO: Loop through the properties and remove anything
                            # greater than 5kb
                            response = analytics.track(
                                user_id=user_id,
                                event=event_name,
                            )

                    asyncio.ensure_future(send_analytics())
            except Exception as e:
                pass
            try:
                return func(*args, **kwargs)
            finally:
                os.environ[TRANSIT_ENV_VAR] = "FALSE"

        return wrapper

    return track


def identify(func: Callable):
    def wrapper(*args, **kwargs):
        try:
            if is_tracking_enabled():
                user_id = args[0].firebase_uid
                region = args[0].region
                traits = {
                    "region": region,
                }
                if user_id is not None:
                    analytics.identify(user_id, traits)
        except Exception as e:
            pass

        return func(*args, **kwargs)

    return wrapper



"""ClusterOps 
"""
# from relevanceai.ops.clusterops.clusterops import ClusterOps
# from relevanceai.ops.clusterops.clusterbase import (
#     ClusterBase,
#     CentroidClusterBase,
#     AdvancedCentroidClusterBase,
# )
# from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel



import warnings
from functools import wraps


def beta(f):
    old_doc = f.__doc__
    f.__doc__ = (
        old_doc
        + """

.. warning::
    This function is currently in beta and is liable to change in the future. We
    recommend not using this in production systems.

    """
    )

    @wraps(f)
    def wrapper(*args, **kwds):
        warnings.warn("This function currently in beta and may change in the future.")
        return f(*args, **kwds)

    return wrapper


def introduced_in_version(version_number):
    def _version(f):
        old_doc = f.__doc__
        new_doc = (
            old_doc
            + f"""

.. note::
    This function was introduced in **{version_number}**.

        """
        )
        f.__doc__ = new_doc

        @wraps(f)
        def wrapper(*args, **kwds):
            return f(*args, **kwds)

        return wrapper

    return _version


def deprecated(version_number: str, message: str = ""):
    def _version(f):
        old_doc = f.__doc__
        new_doc = (
            old_doc
            + f"""

.. note::
    This function has been deprecated as of {version_number}

        """
        )
        f.__doc__ = new_doc

        @wraps(f)
        def wrapper(*args, **kwds):
            DEPRECATION_MESSAGE = (
                f"Deprecated. Revert to versions before {version_number} for function. "
                + message
            )
            warnings.warn(DEPRECATION_MESSAGE, DeprecationWarning)
            return f(*args, **kwds)

        return wrapper

    return _version



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import warnings
import itertools
from itertools import chain
from collections import Counter, defaultdict
from typing import Callable, Dict, List, Optional
from tqdm.auto import tqdm
from relevanceai.package_utils.analytics_funcs import track
from relevanceai.dataset.crud.dataset_write import Write
from relevanceai.unstructured_data.text.base_text_processing import MLStripper
from relevanceai.package_utils.logger import FileLogger
from relevanceai.package_utils.version_decorators import introduced_in_version, beta
from relevanceai.vector_tools.local_nearest_neighbours import (
    NearestNeighbours,
    NEAREST_NEIGHBOURS,
)

# TODO: Separate out operations into different files - cluster/search/dr
class DimensionalityReduction(Write):
    def _run_dr_algorithm(
        self,
        algorithm: str,
        vector_fields: list,
        documents: list,
        alias: str,
        n_components: int = 3,
    ):
        original_name = algorithm
        # Make sure that the letter case does not matter
        algorithm = algorithm.upper()
        if algorithm == "PCA":
            from relevanceai.workflows.dim_reduction_ops.dim_reduction import PCA

            model = PCA()
        elif algorithm == "TSNE":
            from relevanceai.workflows.dim_reduction_ops.dim_reduction import TSNE

            model = TSNE()
        elif algorithm == "UMAP":
            from relevanceai.workflows.dim_reduction_ops.dim_reduction import UMAP

            model = UMAP()
        elif algorithm == "IVIS":
            from relevanceai.workflows.dim_reduction_ops.dim_reduction import Ivis

            model = Ivis()
        else:
            raise ValueError(
                f'"{original_name}" is not a supported '
                "dimensionality reduction algorithm. "
                "Currently, the supported algorithms are: "
                "PCA, TSNE, UMAP, and IVIS"
            )

        print(f"Run {algorithm}...")
        # Returns a list of documents with dr vector
        return model.fit_transform_documents(
            vector_field=vector_fields[0],
            documents=documents,
            alias=alias,
            dims=n_components,
        )

    @track
    def auto_reduce_dimensions(
        self,
        alias: str,
        vector_fields: list,
        filters: Optional[list] = None,
        number_of_documents: Optional[int] = None,
    ):
        """
        Run dimensionality reduction quickly on a dataset on a small number of documents.
        This is useful if you want to quickly see a projection of your dataset.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        Parameters
        ----------
        vector_fields: list
            The vector fields to run dimensionality reduction on
        number_of_documents: int
            The number of documents to get
        algorithm: str
            The algorithm to run. The only supported algorithm is `pca` at this
            current point in time.
        n_components: int
            The number of components

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            df.auto_reduce_dimensions(
                "pca-3",
                ["sample_vector_"],
            )

        """
        if len(vector_fields) > 1:
            raise ValueError("We only support 1 vector field at the moment.")

        dr_args = alias.split("-")

        if len(dr_args) != 2:
            raise ValueError("""Your DR alias should be in the form of `pca-3`.""")

        algorithm = dr_args[0]
        n_components = int(dr_args[1])

        print("Getting documents...")
        if filters is None:
            filters = []
        filters += [
            {
                "field": vf,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vf in vector_fields
        ]

        if number_of_documents is None:
            number_of_documents = self.get_number_of_documents(self.dataset_id, filters)

        documents = self.get_documents(
            select_fields=vector_fields,
            filters=filters,
            number_of_documents=number_of_documents,
        )

        dr_documents = self._run_dr_algorithm(
            algorithm=algorithm,
            vector_fields=vector_fields,
            documents=documents,
            alias=alias,
            n_components=n_components,
        )

        results = self.update_documents(self.dataset_id, dr_documents)

        if n_components == 3:
            projector_url = f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/projector"
            print(f"You can now view your projector at {projector_url}")

        return results

    @track
    def reduce_dimensions(
        self,
        vector_fields: list,
        alias: str,
        number_of_documents: int = 1000,
        algorithm: str = "pca",
        n_components: int = 3,
        filters: Optional[list] = None,
    ):
        """
        Run dimensionality reduction quickly on a dataset on a small number of documents.
        This is useful if you want to quickly see a projection of your dataset.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        Parameters
        ----------
        vector_fields: list
            The vector fields to run dimensionality reduction on
        number_of_documents: int
            The number of documents to get
        algorithm: str
            The algorithm to run. The only supported algorithm is `pca` at this
            current point in time.
        n_components: int
            The number of components

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            df.auto_reduce_dimensions(
                alias="pca-3",
                ["sample_vector_"],
                number_of_documents=1000
            )

        """
        filters = [] if filters is None else filters

        if len(vector_fields) > 1:
            raise ValueError("We only support 1 vector field at the moment.")

        print("Getting documents...")
        filters += [
            {
                "field": vf,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for vf in vector_fields
        ]
        documents = self.get_documents(
            select_fields=vector_fields,
            filters=filters,
            number_of_documents=number_of_documents,
        )

        dr_documents = self._run_dr_algorithm(
            algorithm=algorithm,
            vector_fields=vector_fields,
            documents=documents,
            alias=alias,
            n_components=n_components,
        )

        results = self.update_documents(self.dataset_id, dr_documents)

        return results



import pandas as pd


class JsonShowerMixin:
    def _show_json(
        self, documents, image_fields, audio_fields, highlight_fields, text_fields, **kw
    ):
        from jsonshower import show_json

        if not text_fields:
            text_fields = pd.json_normalize(documents).columns.tolist()
        else:
            text_fields = text_fields
        return show_json(
            documents,
            image_fields=image_fields,
            audio_fields=audio_fields,
            highlight_fields=highlight_fields,
            text_fields=text_fields,
            **kw,
        )



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import traceback
from itertools import chain
from collections import Counter, defaultdict
from typing import Callable, Dict, List, Optional
from tqdm.auto import tqdm
from relevanceai.package_utils.analytics_funcs import track
from relevanceai.dataset.crud.dataset_write import Write
from relevanceai.unstructured_data.text.base_text_processing import MLStripper
from relevanceai.package_utils.logger import FileLogger
from relevanceai.package_utils.version_decorators import (
    introduced_in_version,
    beta,
)
from relevanceai.vector_tools.local_nearest_neighbours import (
    NearestNeighbours,
    NEAREST_NEIGHBOURS,
)


# TODO: Separate out operations into different files - cluster/search/dr


class CommunityDetection(Write):
    @track
    def community_detection(
        self,
        field: str,
        model=None,
        retrieval_kwargs: Optional[dict] = None,
        encode_kwargs: Optional[dict] = None,
        threshold: float = 0.75,
        min_community_size: int = 3,
        init_max_size: int = 1000,
        update_chunksize: int = 100,
        alias: str = "community-detection",
        log_file: str = "logs.txt",
        cluster_field: str = "_cluster_",
    ):
        """
        Performs community detection on a text field.

        Parameters
        ----------
        field: str
            The field over which to find communities. Must be of type "text"
            or "vector".

        model
            A model for computing sentence embeddings.

        retrieval_kwargs: Optional[dict]
            Keyword arguments for `get_documents` call. See respective
            details for argument details.

        encode_kwargs: Optional[dict]
            Keyword arguments for the provide model's `encode` call. See
            respective method for argument details.

        threshold: float
            A lower limit of similarity that determines whether two embeddings
            are similar or not.

        min_community_size: int
            The minimum size of a community. Only communities that are larger
            than this value are returned, and the first element of each
            community is treated as the central point.

        init_max_size: int
            The maximum size of a community. If the corpus is larger than this
            value, that is set to the maximum size.

        Example
        -------

        .. code-block::

            from relevanceai import Client

            client = Client()

            ds = client.Dataset("sample_dataset_id")

            communities = ds.community_detection("sample_text_field")

        """
        if field in self.schema:
            if not (self.schema[field] == "text" or field.endswith("_vector_")):
                raise ValueError("The field must be a 'text' type or a vector")
        else:
            raise ValueError(f"{field} does not exist in the dataset")

        try:
            with FileLogger(log_file):
                from sentence_transformers.util import community_detection
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "community_detection function not found. "
                "Please install sentence-transformers with `python -m "
                "pip install -U sentence-transformers` to install "
                "community_detection."
            )

        field_type = "text" if self.schema[field] == "text" else "vector"

        retrieval_kwargs = {} if retrieval_kwargs is None else retrieval_kwargs
        encode_kwargs = {} if encode_kwargs is None else encode_kwargs

        if model is None and field_type == "text":
            from sentence_transformers import SentenceTransformer

            model = SentenceTransformer("all-MiniLM-L6-v2")
            # encode defaults:
            #  batch_size: int = 32
            #  show_progress_bar: bool = None
            #  output_value: str = 'sentence_embedding'
            #  convert_to_numpy: bool = True
            #  convert_to_Tensor: bool = False
            #  device: str = None
            #  normalize_embeddings: bool = False

        print("Retrieving documents...")
        documents = self.get_all_documents(
            select_fields=[field],
            filters=[
                {
                    "field": field,
                    "filter_type": "exists",
                    "condition": "==",
                    "condition_value": " ",
                }
            ],
            **{
                key: value
                for key, value in retrieval_kwargs.items()
                if key != "select_fields"
            },
        )
        print("Documents retrieved.")

        # lists (i.e. vectors) are unhashable therefore handled by converting
        # them into tuples.
        if field_type == "vector":
            for document in documents:
                try:
                    value = tuple(self.get_field(field, document))
                    self.set_field(field, document, value)
                except KeyError:
                    # If a document is missing a vector, ignore
                    continue

        # Keep track of the fields. Since two documents could have the same
        # field, use a list to keep track of multiples
        element_ids = defaultdict(list)

        # remove duplicates
        elements = set()
        for document in documents:
            try:
                element = self.get_field(field, document)
            except Exception as e:
                # It could be that a document does not have a field or a
                # a vector that other documents in the Dataset has. In that
                # case, ignore.
                traceback.print_exc()
                continue
            elements.add(element)
            element_ids[element].append(document["_id"])

        # Storing a mapping like this is fine because when the values are
        # made a list below, they will be a list in the same order as inserted
        # in the dictionary.
        element_map = {}
        ids_map = {}
        for i, element in enumerate(elements):
            element_map[i] = element
            ids_map[i] = element_ids[element]

        if field_type == "text":
            print("Encoding the corpus...")
            embeddings = model.encode(
                sentences=list(element_map.values()),
                **{
                    key: value
                    for key, value in encode_kwargs.items()
                    if key != "sentences"
                },
            )
            print("Encoding complete.")
        else:
            from numpy import array

            embeddings = array(list(element_map.values()))

        print("Community detection started...")
        init_max_size = min(init_max_size, embeddings.shape[0])
        clusters = community_detection(
            embeddings=embeddings,
            threshold=threshold,
            min_community_size=min_community_size,
            init_max_size=init_max_size,
        )
        print(f"Detected {len(clusters)} communities.")

        # TODO: add centroids for community detection

        print("Updating documents...")
        community_documents = []
        for i, cluster in enumerate(clusters):
            ids = []
            for member in cluster:
                ids.extend(ids_map[member])
            # During initial construction update_where did not accept dict
            # values as valid updates.
            # self.datasets.documents.update_where(
            #    self.dataset_id,
            #    update={
            #        "_cluster_": {field: {"community-detection": f"community-{i+1}"}}
            #    },
            #    filters=[
            #        {
            #            "field": "ids",
            #            "filter_type": "ids",
            #            "condition": "==",
            #            "condition_value": ids,
            #        }
            #    ],
            # )
            for id in ids:
                community_documents.append(
                    {
                        "_id": id,
                        cluster_field: {field: {alias: f"cluster-{i+1}"}},
                    }
                )

            # During initial construction update_where did not accept dict
            # values as valid updates.
            with FileLogger(log_file):
                result = self.datasets.documents.update_where(
                    self.dataset_id,
                    update={"_cluster_": {field: {alias: f"cluster-{i+1}"}}},
                    filters=[
                        {
                            "field": "ids",
                            "filter_type": "ids",
                            "condition": "==",
                            "condition_value": ids,
                        }
                    ],
                )
                print(result)

        print(
            "Build your clustering app here: "
            f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
        )
        # Return a ClusterOps object

        from relevanceai.workflows.cluster_ops.clusterops import ClusterOps

        cluster_ops = ClusterOps(
            # model=model,
            alias=alias,
            dataset_id=self.dataset_id,
            vector_fields=[field],
            cluster_field=cluster_field,
            project=self.project,
            api_key=self.api_key,
            firebase_uid=self.firebase_uid,
        )
        print("Creating centroids...")
        with FileLogger(log_file):
            result = cluster_ops.create_centroids([field])
        print("✅ Uploaded centroids.")
        return cluster_ops



from typing import Optional
from relevanceai.dashboard.dashboard import Dashboard


class Clusters(Dashboard):
    def __init__(self, project: str, api_key: str, deployable_id: str, firebase_uid):
        super().__init__(project, api_key, deployable_id, "cluster", firebase_uid)

    @classmethod
    def create_dashboard(
        cls,
        project: str,
        api_key: str,
        dataset_id: str,
        vector_field: str,
        alias: str,
        share: bool = False,
        firebase_uid: Optional[str] = None,
        **configuration
    ):
        # TODO: if there is no _cluster_ field in schema, create it here?
        application = "cluster"
        application_configuration = {
            "collection_name": dataset_id,
            "type": application,
            "deployable_name": dataset_id,
            "project_id": project,
            application: {
                "alias": alias,
                "vector_field": vector_field,
                **{
                    key: value
                    for key, value in configuration.items()
                    if key not in {"alias", "vector_field"}
                },
            },
        }

        return Dashboard.create_dashboard(
            project,
            api_key,
            dataset_id,
            vector_field,
            application,
            share,
            application_configuration,
            firebase_uid,  # type: ignore
        )
